/* modcmd.c - NeonServ v5.6
 * Copyright (C) 2011-2012  Philipp Kreil (pk910)
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License 
 * along with this program. If not, see <http://www.gnu.org/licenses/>. 
 */

#include "modcmd.h"
#include "IRCEvents.h"
#include "IRCParser.h"
#include "ClientSocket.h"
#include "UserNode.h"
#include "ChanNode.h"
#include "ChanUser.h"
#include "WHOHandler.h"
#include "lang.h"
#include "mysqlConn.h"
#include "DBHelper.h"
#include "EventLogger.h"

struct trigger_callback {
    int botid;
    int module_id;
    trigger_callback_t *func;
    
    struct trigger_callback *next;
};

struct cmd_bot_alias {
    int botid;
    char *alias;
    
    struct cmd_bot_alias *next;
};

struct command_check_user_cache {
    struct ClientSocket *client, *textclient;
    struct UserNode *user;
    struct ChanNode *chan, *sent_chan;
    char **argv;
    int argc;
    char *message, *args_buffer;
    struct cmd_binding *cbind;
};

static struct cmd_binding **cmd_binds;
static struct cmd_function *cmd_functions = NULL;
static struct trigger_callback *trigger_callbacks = NULL;
static struct cmd_bot_alias *bot_aliases = NULL;
static int total_triggered = 0;
int statistics_commands = 0;

static const struct default_language_entry msgtab[] = {
    {"MODCMD_LESS_PARAM_COUNT", "This command requires more parameters."},
    {"MODCMD_CHAN_REQUIRED",    "You must provide the name of a channel that exists and the bot is on."},
    {"MODCMD_AUTH_REQUIRED",    "You need to be authenticated with AuthServ to use this command."},
    {"MODCMD_CHAN_SUSPENDED",   "This channel is currently suspended."},
    {"MODCMD_PRIVILEGED",       "$b%s$b is a privileged command."}, /* {ARGS: "god"} */
    {"MODCMD_PUBCMD",           "Public commands in $b%s$b are restricted."}, /* {ARGS: "#TestChan"} */
    {"MODCMD_ACCESS_DENIED",    "Access denied."},
    {"MODCMD_SUBCOMMANDS",      "Subcommands of %s: %s"}, /* {ARGS: "bot", "ADD, DEL, EDIT"} */
    {"MODCMD_CROSSCHAN",        "You must be in %s (or on its userlist) to use this command."}, /* {ARGS: "#TestChan"} */
    {"MODCMD_UNKNOWN",          "$b%s$b is an unknown command."}, /* {ARGS: "bla"} */
    {NULL, NULL}
};

static int get_binds_index(char first_char) {
    if(tolower(first_char) >= 'a' && tolower(first_char) <= 'z') {
        return tolower(first_char) - 'a';
    }
    return 26;
}

struct ClientSocket* get_botwise_prefered_bot(int botid, int clientid) {
    struct ClientSocket *client, *lowbot = NULL;
    for(client = getBots(SOCKET_FLAG_READY, NULL); client; client = getBots(SOCKET_FLAG_READY, client)) {
        if(client->botid == botid && (botid || clientid == client->clientid)) {
            if((client->flags & SOCKET_FLAG_PREFERRED))
                return client;
            else
                lowbot = client;
        }
    }
    return lowbot;
}

static char* get_channel_trigger(int botid, int clientid, struct ChanNode *chan) {
    struct trigger_cache *trigger;
    for(trigger = chan->trigger; trigger; trigger = trigger->next) {
        if(trigger->botid == botid && (botid || trigger->clientid == clientid))
            return trigger->trigger;
    }
    struct trigger_callback *cb;
    for(cb = trigger_callbacks; cb; cb = cb->next) {
        if(cb->botid == botid)
            break;
    }
    char triggerStr[TRIGGERLEN];
    if(cb)
        cb->func(clientid, chan, triggerStr);
    else
        triggerStr[0] = '\0';
    trigger = malloc(sizeof(*trigger));
    if (!trigger) {
        perror("malloc() failed");
        return 0;
    }
    trigger->botid = botid;
    trigger->clientid = clientid;
    trigger->trigger = (triggerStr[0] ? strdup(triggerStr) : NULL);
    trigger->next = chan->trigger;
    chan->trigger = trigger;
    return trigger->trigger;
}

static void handle_command_async(struct ClientSocket *client, struct ClientSocket *textclient, struct UserNode *user, struct ChanNode *chan, struct ChanNode *sent_chan, struct cmd_binding *cbind, char **argv, int argc);

static USERAUTH_CALLBACK(command_checked_auth) {
    struct command_check_user_cache *cache = data;
    if(user)
        handle_command_async(cache->client, cache->textclient, user, cache->chan, cache->sent_chan, cache->cbind, cache->argv, cache->argc);
    free(cache->message);
    if(cache->args_buffer)
        free(cache->args_buffer);
    free(cache->argv);
    free(cache);
}

static CMD_BIND(modcmd_linker) {
    //fake command for subcommands
    //just empty
}

static struct cmd_binding *modcmd_linker_command(struct ClientSocket *client, struct ClientSocket *textclient, struct UserNode *user, struct cmd_binding *cbind, int bind_index, char **args_ptr) {
    //links subcommands
    char command[MAXLEN];
    struct cmd_binding *parent_bind = cbind;
    char *args = *args_ptr;
    if(args) {
        char *subcmd = args;
        args = strstr(args, " ");
        if(args) {
            *args = '\0';
            args++;
        }
        sprintf(command, "%s %s", parent_bind->cmd, subcmd);
        for(cbind = cmd_binds[bind_index]; cbind; cbind = cbind->next) {
            if(cbind->botid == client->botid && (cbind->botid || cbind->clientid == client->clientid) && stricmp(cbind->cmd, command) == 0)
                break;
        }
        *args_ptr = args;
        if(cbind && cbind->func->func == modcmd_linker) {
            return modcmd_linker_command(client, textclient, user, cbind, bind_index, args_ptr);
        }
        return cbind;
    } else {
        //list all sub commands
        int commandlen = sprintf(command, "%s ", parent_bind->cmd);
        char subcommands[MAXLEN];
        int subcompos = 0;
        for(cbind = cmd_binds[bind_index]; cbind; cbind = cbind->next) {
            if(cbind->botid == client->botid && (cbind->botid || cbind->clientid == client->clientid) && stricmplen(cbind->cmd, command, commandlen) == 0) {
                if(strstr(cbind->cmd + commandlen, " ")) continue; //ignore if there is another space in the command (sub-sub-command :D)
                subcompos += sprintf(subcommands + subcompos, (subcompos ? ", %s" : "%s"), cbind->cmd + commandlen);
            }
        }
        reply(textclient, user, "MODCMD_SUBCOMMANDS", parent_bind->cmd, (subcompos ? subcommands : "none"));
        return NULL;
    }
}

static struct cmd_binding *modcmd_sub_linker_command(struct ClientSocket *client, struct ClientSocket *textclient, struct UserNode *user, struct cmd_binding *cbind, int bind_index, char **args_ptr) {
    //links subcommands
    char command[MAXLEN];
    struct cmd_binding *parent_bind = cbind;
    char *args = *args_ptr;
    if(args) {
        char *subcmd = args;
        args = strchr(args, ' ');
        if(args) {
            *args = '\0';
            args++;
        }
        sprintf(command, "%s %s", parent_bind->cmd, subcmd);
        for(cbind = cmd_binds[bind_index]; cbind; cbind = cbind->next) {
            if(cbind->botid == client->botid && (cbind->botid || cbind->clientid == client->clientid) && stricmp(cbind->cmd, command) == 0)
                break;
        }
        *args_ptr = args;
        if(cbind && cbind->func->func == modcmd_linker) {
            return modcmd_linker_command(client, textclient, user, cbind, bind_index, args_ptr);
        }
        return (cbind ? cbind : parent_bind);
    } else
        return parent_bind;
}

static void handle_command(struct ClientSocket *client, struct UserNode *user, struct ChanNode *chan, char *message) {
    struct ChanNode *sent_chan = chan;
    if(message[0] == '#') {
        char *chanName = message;
        message = strstr(message, " ");
        if(!message) return;
        *message = '\0';
        message++;
        struct ChanNode *chan2 = getChanByName(chanName);
        if(chan2)
            chan = chan2;
    }
    message = strdup(message);
    int bind_index = get_binds_index(message[0]);
    char *args = strchr(message, ' ');
    char *args_buffer = NULL; //we need this to save a possible pointer to a allocation we need to free
    if(args) {
        *args = '\0';
        args++;
    }
    struct cmd_binding *cbind;
    int found_cmd = 0;
    for(cbind = cmd_binds[bind_index]; cbind; cbind = cbind->next) {
        if(cbind->botid == client->botid && (cbind->botid || cbind->clientid == client->clientid) && stricmp(cbind->cmd, message) == 0) {
            found_cmd = 1;
            //get a text bot
            struct ClientSocket *textclient = get_botwise_prefered_bot(client->botid, (client->botid == 0 ? client->clientid : 0));
            if(cbind->func->func == modcmd_linker) {
                cbind = modcmd_linker_command(client, textclient, user, cbind, bind_index, &args);
                if(cbind == NULL) break;
            } else if(cbind->flags & CMDFLAG_SUB_LINKER)
                cbind = modcmd_sub_linker_command(client, textclient, user, cbind, bind_index, &args);
            if(statistics_enabled)
                statistics_commands++;
            total_triggered++;
            cbind->triggered++;
            if((BIND_FLAGS(cbind) & CMDFLAG_FUNCMD)) {
                if(!sent_chan)
                    break;
                chan = sent_chan;
            }
            //parse the arguments...
            char *arga[MAXNUMPARAMS];
            char **argv;
            int argc = 0;
            int escape = 0;
            int offset = 0;
            if(args) {
                while(*args) {
                    //skip leading spaces
                    while (*args == ' ')
                        *args++ = 0;
                    arga[argc++] = args;
                    if (argc >= MAXNUMPARAMS)
                        break;
                    while ((escape || *args != ' ') && *args) {
                        if((BIND_FLAGS(cbind) & CMDFLAG_ESCAPE_ARGS) && *args == '\\') {
                            escape = 1;
                            offset++;
                        } else if(escape)
                            escape = 0;
                        if(!escape && offset) {
                            args[0 - offset] = args[0];
                        }
                        args++;
                        
                    }
                    if(offset) {
                        args[0-offset] = '\0';
                        offset = 0;
                    }
                }
            }
            argv = arga;
            if(cbind->paramcount) {
                //userdefined parameters...
                args_buffer = malloc(MAXLEN * 2 * sizeof(*args_buffer));
                int args_pos = 0;
                char *uargs[MAXNUMPARAMS];
                int uargc = 0;
                char *b, *c;
                int i;
                int allargs, argi;
                for(i = 0; i < cbind->paramcount; i++) {
                    b = cbind->parameters[i];
                    if(b[0] == '%') {
                        b++;
                        c = b;
                        while(*c && *c != ' ') {
                            if(*c == '|') break;
                            c++;
                        }
                        if(!*c || *c == ' ') c = NULL;
                        else {
                            *c = '\0';
                            c++;
                        }
                        if(b[strlen(b)-1] == '-') {
                            allargs = strlen(b)-1;
                            b[allargs] = '\0';
                            argi = atoi(b);
                            b[allargs] = '-';
                            allargs = 1;
                        } else if(b[strlen(b)-1] == '+') {
                            allargs = strlen(b)-1;
                            b[allargs] = '\0';
                            argi = atoi(b);
                            b[allargs] = '+';
                            allargs = 2;
                        } else {
                            allargs = 0;
                            argi = atoi(b);
                        }
                        if(argi > 0) {
                            if(argi <= argc) {
                                uargs[uargc++] = args_buffer + args_pos;
                                if(allargs == 0) {
                                    args_pos += sprintf(args_buffer + args_pos, "%s", argv[argi-1]) + 1;
                                } else if(allargs == 1) {
                                    args_pos += sprintf(args_buffer + args_pos, "%s", argv[argi-1]) + 1;
                                    for(argi++; argi <= argc; argi++) {
                                        uargs[uargc++] = args_buffer + args_pos;
                                        args_pos += sprintf(args_buffer + args_pos, "%s", argv[argi-1]) + 1;
                                    }
                                } else if(allargs == 2) {
                                    for(;argi <= argc; argi++) {
                                        args_pos += sprintf(args_buffer + args_pos, (allargs ? "%s" : " %s"), argv[argi-1]);
                                        allargs = 0;
                                    }
                                    args_pos++;
                                }
                            } else if((BIND_FLAGS(cbind) & CMDFLAG_EMPTY_ARGS)) {
                                uargs[uargc++] = args_buffer + args_pos;
                                args_buffer[args_pos++] = '\0';
                            } else if(c) {
                                uargs[uargc++] = args_buffer + args_pos;
                                args_pos += sprintf(args_buffer + args_pos, "%s", c) + 1;
                            }
                        } else if(!strcmp(b, "c")) {
                            uargs[uargc++] = args_buffer + args_pos;
                            args_pos += sprintf(args_buffer + args_pos, "%s", (chan ? chan->name : "")) + 1;
                        } else if(!strcmp(b, "n")) {
                            uargs[uargc++] = args_buffer + args_pos;
                            args_pos += sprintf(args_buffer + args_pos, "%s", user->nick) + 1;
                        }
                        if(c) c[-1] = '|'; //reset \0 to |
                    } else {
                        uargs[uargc++] = args_buffer + args_pos;
                        args_pos += sprintf(args_buffer + args_pos, "%s", b) + 1;
                    }
                }
                argv = uargs;
                argc = uargc;
            }
            if(argc != 0 && argv[0][0] == '#' && !(BIND_FLAGS(cbind) & CMDFLAG_CHAN_PARAM)) {
                struct ChanNode *chan2 = getChanByName(argv[0]);
                if(chan2) {
                    argv += 1;
                    argc -= 1;
                    chan = chan2;
                }
            }
            if(argc < cbind->func->paramcount) {
                reply(textclient, user, "MODCMD_LESS_PARAM_COUNT");
                break;
            }
            if((BIND_FLAGS(cbind) & CMDFLAG_REQUIRE_CHAN) && !chan) {
                reply(textclient, user, "MODCMD_CHAN_REQUIRED");
                break;
            }
            if(((BIND_FLAGS(cbind) & CMDFLAG_CHECK_AUTH) || (chan && chan != sent_chan && !isUserOnChan(user, chan))) && !(user->flags & USERFLAG_ISAUTHED)) {
                //check auth...
                struct command_check_user_cache *data = malloc(sizeof(*data));
                char **temp_argv = malloc(argc*sizeof(*temp_argv));
                if (!data || !temp_argv) {
                    perror("malloc() failed");
                    break;
                }
                memcpy(temp_argv, argv, argc*sizeof(*temp_argv));
                data->argv = temp_argv;
                data->argc = argc;
                data->client = client;
                data->user = user;
                data->chan = chan;
                data->sent_chan = sent_chan;
                data->message = message;
                data->args_buffer = args_buffer;
                data->cbind = cbind;
                data->textclient = textclient;
                get_userauth(user, 0, command_checked_auth, data);
                return;
            } else
                handle_command_async(client, textclient, user, chan, sent_chan, cbind, argv, argc);
            break;
        }
    }
    if(!found_cmd && !sent_chan && !(client->flags & SOCKET_FLAG_SILENT))
        reply(get_botwise_prefered_bot(client->botid, (client->botid == 0 ? client->clientid : 0)), user, "MODCMD_UNKNOWN", message);
    free(message);
    if(args_buffer)
        free(args_buffer);
}

static void handle_command_async(struct ClientSocket *client, struct ClientSocket *textclient, struct UserNode *user, struct ChanNode *chan, struct ChanNode *sent_chan, struct cmd_binding *cbind, char **argv, int argc) {
    MYSQL_RES *res;
    MYSQL_ROW row;
    int uaccess;
    char requested_uaccess = 0;
    int eventflags = (BIND_FLAGS(cbind) & (CMDFLAG_LOG | CMDFLAG_OPLOG));
    if((BIND_FLAGS(cbind) & CMDFLAG_REQUIRE_AUTH) && !(user->flags & USERFLAG_ISAUTHED)) {
        reply(textclient, user, "MODCMD_AUTH_REQUIRED");
        return;
    }
    if(chan && sent_chan != chan && (BIND_FLAGS(cbind) & CMDFLAG_NO_CROSSCHAN) && !isUserOnChan(user, chan)) {
        char user_in_chan = 0;
        if((user->flags & USERFLAG_ISAUTHED)) {
            //maybe there's another user authed to user->auth on the channel...
            struct ChanUser *cchanuser;
            for(cchanuser = getChannelUsers(chan, NULL); cchanuser; cchanuser = getChannelUsers(chan, cchanuser)) {
                if((cchanuser->user->flags & USERFLAG_ISAUTHED) && !stricmp(user->auth, cchanuser->user->auth)) {
                    user_in_chan = 1;
                    break;
                }
            }
        }
        if(!user_in_chan) {
            //check if we are allowed to execute commands in this channel
            requested_uaccess = 1;
            uaccess = getChannelAccess(user, chan);
            if(!uaccess) {
                if(isGodMode(user)) {
                    eventflags |= CMDFLAG_OPLOG;
                } else {
                    reply(textclient, user, "MODCMD_CROSSCHAN", chan->name);
                    return;
                }
            }
        }
    }
    if(sent_chan && sent_chan != chan) {
        //check pubcmd of this channel
        printf_mysql_query("SELECT `channel_pubcmd` FROM `channels` WHERE `channel_name` = '%s'", escape_string(sent_chan->name));
        res = mysql_use();
        if ((row = mysql_fetch_row(res)) != NULL) {
            int saccess = getChannelAccess(user, sent_chan);
            if(row[0] && saccess < atoi(row[0]) && !isGodMode(user)) { //NOTE: HARDCODED DEFAULT: pubcmd = 0
                reply(textclient, user, "MODCMD_PUBCMD", sent_chan->name);
                return;
            }
        }
    }
    int global_access = ((cbind->flags & CMDFLAG_OVERRIDE_GLOBAL_ACCESS) ? cbind->global_access : cbind->func->global_access);
    if(global_access > 0) {
        int user_global_access = 0;
        printf_mysql_query("SELECT `user_access` FROM `users` WHERE `user_user` = '%s'", escape_string(user->auth));
        res = mysql_use();
        if ((row = mysql_fetch_row(res)) != NULL) {
            user_global_access = atoi(row[0]);
        }
        if(user_global_access < global_access) {
            if(!user_global_access)
                reply(textclient, user, "MODCMD_PRIVILEGED", cbind->cmd);
            else
                reply(textclient, user, "MODCMD_ACCESS_DENIED");
            return;
        }
    }
    if((BIND_FLAGS(cbind) & CMDFLAG_REGISTERED_CHAN)) {
        MYSQL_ROW defaults = NULL;
        char access_list[256];
        int access_pos = 0;
        int access_count = 0;
        int minaccess = 0;
        char *str_a, *str_b = cbind->func->channel_access, *str_c;
        if(cbind->flags & CMDFLAG_OVERRIDE_CHANNEL_ACCESS)
            str_b = cbind->channel_access;
        access_list[0] = '\0';
        if(str_b) {
            struct ChanUser *chanuser = getChanUser(user, chan);
            str_c = strdup(str_b);
            str_b = str_c;
            while((str_a = str_b)) {
                str_b = strstr(str_a, ",");
                if(str_b) {
                    *str_b = '\0';
                    str_b++;
                }
                if(*str_a == '@' || *str_a == '+') {
                    //privs can override this access requirement
                    int priv = 0;
                    if(*str_a == '@') priv = CHANUSERFLAG_OPPED;
                    else if(*str_a == '%') priv = CHANUSERFLAG_HALFOPPED;
                    else if(*str_a == '+') priv = CHANUSERFLAG_VOICED;
                    if(chanuser && (chanuser->flags & priv)) continue;
                    str_a++;
                }
                if(*str_a == '#') {
                    str_a++;
                    access_pos += sprintf(access_list+access_pos, ", `%s`", str_a);
                    access_count++;
                } else {
                    if(atoi(str_a) > minaccess)
                        minaccess = atoi(str_a);
                }
            }
            free(str_c);
        }
        if(!(chan->flags & CHANFLAG_REQUESTED_CHANINFO) || (sent_chan && sent_chan == chan) || access_count || minaccess) {
            printf_mysql_query("SELECT `channel_id`, `channel_pubcmd` %s FROM `channels` WHERE `channel_name` = '%s'", access_list, escape_string(chan->name));
            res = mysql_use();
            if ((row = mysql_fetch_row(res)) != NULL) {
                chan->flags |= CHANFLAG_CHAN_REGISTERED;
                chan->channel_id = atoi(row[0]);
                if((sent_chan && sent_chan == chan) || access_count || minaccess) {
                    if(!requested_uaccess) uaccess = getChannelAccess(user, chan);
                    if(uaccess < minaccess && isGodMode(user)) {
                        eventflags |= CMDFLAG_OPLOG;
                    } else if(uaccess < minaccess) {
                        //ACCESS DENIED
                        reply(textclient, user, "MODCMD_ACCESS_DENIED");
                        return;
                    }
                    if(!row[1] && !defaults) {
                        printf_mysql_query("SELECT `channel_id`, `channel_pubcmd` %s FROM `channels` WHERE `channel_name` = 'defaults'", access_list);
                        defaults = mysql_fetch_row(mysql_use());
                    }
                    if(sent_chan && (sent_chan == chan) && uaccess < (row[1] ? atoi(row[1]) : atoi(defaults[1]))) {
                        if(isGodMode(user)) {
                            eventflags |= CMDFLAG_OPLOG;
                        } else {
                            //PUBCMD
                            reply(textclient, user, "MODCMD_PUBCMD", chan->name);
                            return;
                        }
                    }
                    int i;
                    for(i = 0; i < access_count; i++) {
                        if(!row[2+i] && !defaults) {
                            printf_mysql_query("SELECT `channel_id`, `channel_pubcmd` %s FROM `channels` WHERE `channel_name` = 'defaults'", access_list);
                            defaults = mysql_fetch_row(mysql_use());
                        }
                        if(uaccess < (row[2+i] ? atoi(row[2+i]) : atoi(defaults[2+i]))) {
                            if(isGodMode(user)) {
                                eventflags |= CMDFLAG_OPLOG;
                            } else {
                                reply(textclient, user, "MODCMD_ACCESS_DENIED");
                                return;
                            }
                        }
                    }
                }
            }
            chan->flags |= CHANFLAG_REQUESTED_CHANINFO;
        }
        if(!(chan->flags & CHANFLAG_CHAN_REGISTERED)) {
            reply(textclient, user, "MODCMD_CHAN_REQUIRED");
            return;
        }
        printf_mysql_query("SELECT `botid`, `suspended` FROM `bot_channels` LEFT JOIN `bots` ON `bot_channels`.`botid` = `bots`.`id` WHERE `chanid` = '%d' AND `botclass` = '%d'", chan->channel_id, client->botid);
        res = mysql_use();
        if ((row = mysql_fetch_row(res)) == NULL) {
            reply(textclient, user, "MODCMD_CHAN_REQUIRED");
            return;
        } else if(!strcmp(row[1], "1")) {
            reply(textclient, user, "MODCMD_CHAN_SUSPENDED");
            return;
        }
    }
    if((BIND_FLAGS(cbind) & CMDFLAG_REQUIRE_GOD) && !isGodMode(user)) {
        reply(textclient, user, "MODCMD_PRIVILEGED", cbind->cmd);
        return;
    }
    struct Event *event = createEvent(client, user, chan, cbind, argv, argc, eventflags);
    cbind->func->func(client, textclient, user, chan, argv, argc, event);
}

static void got_chanmsg(struct UserNode *user, struct ChanNode *chan, char *message) {
    fd_set fds, fds2;
    char *trigger;
    struct ClientSocket *client;
    FD_ZERO(&fds);
    FD_ZERO(&fds2);
    got_chanmsg_loop1:
    for(client = getBots(SOCKET_FLAG_READY, NULL); client; client = getBots(SOCKET_FLAG_READY, client)) {
        if(isUserOnChan(client->user, chan) && (client->flags & SOCKET_FLAG_PREFERRED) && ((client->botid == 0 && !FD_ISSET(client->clientid, &fds)) || (client->botid && !FD_ISSET(client->botid, &fds2))))  {
            FD_SET(client->clientid, &fds);
            FD_SET(client->botid, &fds2);
            trigger = get_channel_trigger(client->botid, client->clientid, chan);
            if(trigger && stricmplen(message, trigger, strlen(trigger)) == 0) {
                handle_command(client, user, chan, message + strlen(trigger));
                goto got_chanmsg_loop1; //Thats really really bad, i know... But we can't count on the "getBots" list anymore after executing a command
            }
        }
    }
    got_chanmsg_loop2:
    for(client = getBots(SOCKET_FLAG_READY, NULL); client; client = getBots(SOCKET_FLAG_READY, client)) {
        if(isUserOnChan(client->user, chan) && ((client->botid == 0 && !FD_ISSET(client->clientid, &fds)) || (client->botid && !FD_ISSET(client->botid, &fds2)))) {
            FD_SET(client->clientid, &fds);
            FD_SET(client->botid, &fds2);
            trigger = get_channel_trigger(client->botid, client->clientid, chan);
            if(trigger && stricmplen(message, trigger, strlen(trigger)) == 0) {
                handle_command(client, user, chan, message + strlen(trigger));
                goto got_chanmsg_loop2; //Thats really really bad, i know... But we can't count on the "getBots" list anymore after executing a command
            }
        }
    }
}

static void got_privmsg(struct UserNode *user, struct UserNode *target, char *message) {
    struct ClientSocket *client;
    for(client = getBots(SOCKET_FLAG_READY, NULL); client; client = getBots(SOCKET_FLAG_READY, client)) {
        if(client->user == target) {
            handle_command(client, user, NULL, message);
        }
    }
}

int register_command(int botid, char *name, int module_id, cmd_bind_t *func, int paramcount, char *channel_access, int global_access, unsigned int flags) {
    struct cmd_function *cmdfunc;
    for(cmdfunc = cmd_functions; cmdfunc; cmdfunc = cmdfunc->next) {
        if((cmdfunc->botid == botid || cmdfunc->botid == 0) && !stricmp(cmdfunc->name, name))
            return 0;
    }
    cmdfunc = malloc(sizeof(*cmdfunc));
    if (!cmdfunc) {
        perror("malloc() failed");
        return 0;
    }
    cmdfunc->botid = botid;
    cmdfunc->name = strdup(name);
    cmdfunc->module_id = module_id;
    cmdfunc->func = func;
    cmdfunc->flags = flags;
    cmdfunc->paramcount = paramcount;
    cmdfunc->channel_access = channel_access;
    cmdfunc->global_access = global_access;
    cmdfunc->next = cmd_functions;
    cmd_functions = cmdfunc;
    return 1;
}

int set_trigger_callback(int botid, int module_id, trigger_callback_t *func) {
    static struct trigger_callback *cb = NULL;
    for(cb = trigger_callbacks; cb; cb = cb->next) {
        if(cb->botid == botid)
            break;
    }
    if(!cb) {
        cb = malloc(sizeof(*cb));
        if (!cb) {
            perror("malloc() failed");
            return 0;
        }
        cb->botid = botid;
        cb->next = trigger_callbacks;
        trigger_callbacks = cb;
    }
    cb->module_id = module_id;
    cb->func = func;
    return 1;
}

int flush_trigger_cache(int botid, int clientid) {
    struct ChanNode *chan;
    struct trigger_cache *trigger, *last;
    for(chan = getAllChans(NULL); chan; chan = getAllChans(chan)) {
        last = NULL;
        for(trigger = chan->trigger; trigger; trigger = trigger->next) {
            if(trigger->botid == botid && (botid || trigger->clientid == clientid)) {
                if(last)
                    last->next = trigger->next;
                else
                    chan->trigger = trigger->next;
                free(trigger->trigger);
                free(trigger);
                break;
            } else
                last = trigger;
        }
    }
    return 1;
}

int changeBotwiseChannelTrigger(int botid, int clientid, struct ChanNode *chan, char *new_trigger) {
    struct trigger_cache *trigger;
    for(trigger = chan->trigger; trigger; trigger = trigger->next) {
        if(trigger->botid == botid && (botid || trigger->clientid == clientid)) {
            free(trigger->trigger);
            trigger->trigger = strdup(new_trigger);
            return 1;
        }
    }
    return 0;
}

int bind_botwise_cmd_to_function(int botid, int clientid, char *cmd, struct cmd_function *func) {
    int bind_index = get_binds_index(cmd[0]);
    struct cmd_binding *cbind;
    for(cbind = cmd_binds[bind_index]; cbind; cbind = cbind->next) {
        if(((botid && cbind->botid == botid) || (botid == 0 && clientid == cbind->clientid)) && !stricmp(cbind->cmd, cmd))
            return 0;
    }
    cbind = malloc(sizeof(*cbind));
    if (!cbind) {
        perror("malloc() failed");
        return 0;
    }
    cbind->botid = botid;
    cbind->clientid = clientid;
    cbind->cmd = strdup(cmd);
    cbind->func = func;
    cbind->paramcount = 0;
    cbind->global_access = 0;
    cbind->channel_access = NULL;
    cbind->flags = 0;
    cbind->triggered = 0;
    cbind->next = cmd_binds[bind_index];
    cmd_binds[bind_index] = cbind;
    return 1;
}

int bind_botwise_cmd_to_command(int botid, int clientid, char *cmd, char *func) {
    struct cmd_function *cmdfunc;
    int fbotid = botid;
    char *c;
    if((c = strstr(func, "."))) {
        *c = '\0';
        struct cmd_bot_alias *botalias;
        for(botalias = bot_aliases; botalias; botalias = botalias->next) {
            if(!stricmp(botalias->alias, func)) {
                fbotid = botalias->botid;
                break;
            }
        }
        *c = '.';
        func = c+1;
    }
    for(cmdfunc = cmd_functions; cmdfunc; cmdfunc = cmdfunc->next) {
        if((cmdfunc->botid == fbotid || cmdfunc->botid == 0) && !stricmp(cmdfunc->name, func))
            break;
    }
    if(!cmdfunc) return 0;
    int bind_index = get_binds_index(cmd[0]);
    struct cmd_binding *cbind;
    for(cbind = cmd_binds[bind_index]; cbind; cbind = cbind->next) {
        if(((botid && cbind->botid == botid) || (botid == 0 && clientid == cbind->clientid)) && !stricmp(cbind->cmd, cmd))
            return 0;
    }
    cbind = malloc(sizeof(*cbind));
    if (!cbind) {
        perror("malloc() failed");
        return 0;
    }
    cbind->botid = botid;
    cbind->clientid = clientid;
    cbind->cmd = strdup(cmd);
    cbind->func = cmdfunc;
    cbind->next = cmd_binds[bind_index];
    cbind->paramcount = 0;
    cbind->global_access = 0;
    cbind->channel_access = NULL;
    cbind->flags = 0;
    cbind->triggered = 0;
    cmd_binds[bind_index] = cbind;
    return 1;
}

int unbind_botwise_cmd(int botid, int clientid, char *cmd) {
    int bind_index = get_binds_index(cmd[0]);
    struct cmd_binding *cbind, *last = NULL;
    for(cbind = cmd_binds[bind_index]; cbind; cbind = cbind->next) {
        if(cbind->botid == botid && (botid || clientid == cbind->clientid) && !stricmp(cbind->cmd, cmd)) {
            if(last)
                last->next = cbind->next;
            else
                cmd_binds[bind_index] = cbind->next;
            free(cbind->cmd);
            if(cbind->paramcount) {
                int i;
                for(i = 0; i < cbind->paramcount; i++)
                    free(cbind->parameters[i]);
            }
            if(cbind->channel_access)
                free(cbind->channel_access);
            free(cbind);
            return 1;
        } else
            last = cbind;
    }
    return 0;
}

int unbind_botwise_allcmd(int botid, int clientid) {
    int i;
    for(i = 0; i < 27; i++) {
        struct cmd_binding *cbind, *next, *last = NULL;
        for(cbind = cmd_binds[i]; cbind; cbind = next) {
            next = cbind->next;
            if((botid == 0 && clientid == cbind->clientid) || (botid && botid == cbind->botid)) {
                if(last)
                    last->next = cbind->next;
                else
                    cmd_binds[i] = cbind->next;
                free(cbind->cmd);
                if(cbind->paramcount) {
                    int j;
                    for(j = 0; j < cbind->paramcount; j++)
                        free(cbind->parameters[j]);
                }
                if(cbind->channel_access)
                    free(cbind->channel_access);
                free(cbind);
            } else
                last = cbind;
        }
    }
    return 1;
}

struct cmd_function *find_cmd_function(int botid, char *name) {
    struct cmd_function *cmdfunc;
    char *c;
    if((c = strstr(name, "."))) {
        *c = '\0';
        struct cmd_bot_alias *botalias;
        for(botalias = bot_aliases; botalias; botalias = botalias->next) {
            if(!stricmp(botalias->alias, name)) {
                botid = botalias->botid;
                break;
            }
        }
        *c = '.';
        name = c+1;
    }
    for(cmdfunc = cmd_functions; cmdfunc; cmdfunc = cmdfunc->next) {
        if((cmdfunc->botid == botid || cmdfunc->botid == 0) && stricmp(cmdfunc->name, name) == 0)
            break;
    }
    return cmdfunc;
}

void init_modcmd() {
    cmd_binds = calloc(27, sizeof(*cmd_binds));
    bind_chanmsg(got_chanmsg, 0);
    bind_privmsg(got_privmsg, 0);
    register_default_language_table(msgtab);
    register_command(0, "linker", 0, modcmd_linker, 0, 0, 0, 0); //fake command for subcommands
}

void free_modcmd() {
    int i;
    for(i = 0; i < 27; i++) {
        struct cmd_binding *cbind, *next;
        for(cbind = cmd_binds[i]; cbind; cbind = next) {
            next = cbind->next;
            free(cbind->cmd);
            if(cbind->paramcount) {
                int j;
                for(j = 0; j < cbind->paramcount; j++)
                    free(cbind->parameters[j]);
            }
            if(cbind->channel_access)
                free(cbind->channel_access);
            free(cbind);
        }
    }
    free(cmd_binds);
    struct cmd_function *cmdfunct, *next;
    for(cmdfunct = cmd_functions; cmdfunct; cmdfunct = next) {
        next = cmdfunct->next;
        free(cmdfunct->name);
        free(cmdfunct);
    }
    struct trigger_callback *cb, *next_cb;
    for(cb = trigger_callbacks; cb; cb = next_cb) {
        next_cb = cb->next;
        free(next_cb);
    }
    struct cmd_bot_alias *botalias, *next_botalias;
    for(botalias = bot_aliases; botalias; botalias = next_botalias) {
        next_botalias = botalias->next;
        free(botalias->alias);
        free(botalias);
    }
    cmd_functions = NULL;
    trigger_callbacks = NULL;
    bot_aliases = NULL;
}

void bind_botwise_set_parameters(int botid, int clientid, char *cmd, char *parameters) {
    int bind_index = get_binds_index(cmd[0]);
    struct cmd_binding *cbind;
    for(cbind = cmd_binds[bind_index]; cbind; cbind = cbind->next) {
        if(cbind->botid == botid && (botid || clientid == cbind->clientid) && !stricmp(cbind->cmd, cmd)) {
            if(cbind->paramcount) {
                int i;
                for(i = 0; i < cbind->paramcount; i++)
                    free(cbind->parameters[i]);
                cbind->paramcount = 0;
            }
            if(parameters) {
                char *a, *b = parameters;
                do {
                    a = strstr(b, " ");
                    if(a) *a = '\0';
                    cbind->parameters[cbind->paramcount++] = strdup(b);
                    if(a) b = a+1;
                } while(a);
            }
            return;
        }
    }
}

void bind_botwise_set_global_access(int botid, int clientid, char *cmd, int gaccess) {
    int bind_index = get_binds_index(cmd[0]);
    struct cmd_binding *cbind;
    for(cbind = cmd_binds[bind_index]; cbind; cbind = cbind->next) {
        if(cbind->botid == botid && (botid || clientid == cbind->clientid) && !stricmp(cbind->cmd, cmd)) {
            if(gaccess > -1) {
                cbind->global_access = gaccess;
                cbind->flags |= CMDFLAG_OVERRIDE_GLOBAL_ACCESS;
            } else {
                cbind->flags &= ~CMDFLAG_OVERRIDE_GLOBAL_ACCESS;
            }
            return;
        }
    }
}

void bind_botwise_set_channel_access(int botid, int clientid, char *cmd, char *chanaccess) {
    int bind_index = get_binds_index(cmd[0]);
    struct cmd_binding *cbind;
    for(cbind = cmd_binds[bind_index]; cbind; cbind = cbind->next) {
        if(cbind->botid == botid && (botid || clientid == cbind->clientid) && !stricmp(cbind->cmd, cmd)) {
            if(cbind->channel_access)
                free(cbind->channel_access);
            if(chanaccess) {
                cbind->channel_access = strdup(chanaccess);
                cbind->flags |= CMDFLAG_OVERRIDE_CHANNEL_ACCESS;
            } else {
                cbind->channel_access = NULL;
                cbind->flags &= ~CMDFLAG_OVERRIDE_CHANNEL_ACCESS;
            }
            return;
        }
    }
}

void bind_botwise_set_bind_flags(int botid, int clientid, char *cmd, unsigned int flags) {
    int bind_index = get_binds_index(cmd[0]);
    struct cmd_binding *cbind;
    for(cbind = cmd_binds[bind_index]; cbind; cbind = cbind->next) {
        if(cbind->botid == botid && (botid || clientid == cbind->clientid) && !stricmp(cbind->cmd, cmd)) {
            cbind->flags |= flags;
            return;
        }
    }
}

struct cmd_binding *find_botwise_cmd_binding(int botid, int clientid, char *cmd) {
    int bind_index = get_binds_index(cmd[0]);
    struct cmd_binding *cbind;
    for(cbind = cmd_binds[bind_index]; cbind; cbind = cbind->next) {
        if(cbind->botid == botid && (botid || clientid == cbind->clientid) && !stricmp(cbind->cmd, cmd)) {
            return cbind;
        }
    }
    return NULL;
}

void bind_botwise_unbound_required_functions(int botid, int clientid) {
    struct cmd_function *cmdfunc;
    int i, found;
    struct cmd_binding *cbind;
    for(cmdfunc = cmd_functions; cmdfunc; cmdfunc = cmdfunc->next) {
        if((cmdfunc->flags & CMDFLAG_REQUIRED)) {
            found = 0;
            for(i = 0; i < 27; i++) {
                for(cbind = cmd_binds[i]; cbind; cbind = cbind->next) {
                    if(cbind->botid == botid && (botid || clientid == cbind->clientid) && cbind->func == cmdfunc) {
                        found = 1;
                        break;
                    }
                }
                if(found)
                    break;
            }
            if(!found && bind_botwise_cmd_to_function(botid, clientid, cmdfunc->name, cmdfunc)) {
                cbind = find_botwise_cmd_binding(botid, clientid, cmdfunc->name);
                cbind->flags |= CMDFLAG_TEMPONARY_BIND;
            }
        }
    }
}

void register_command_alias(int botid, char *alias) {
    struct cmd_bot_alias *botalias;
    for(botalias = bot_aliases; botalias; botalias = botalias->next) {
        if(!stricmp(botalias->alias, alias))
            return;
    }
    botalias = malloc(sizeof(*botalias));
    if (!botalias) {
        perror("malloc() failed");
        return;
    }
    botalias->botid = botid;
    botalias->alias = strdup(alias);
    botalias->next = bot_aliases;
    bot_aliases = botalias;
}

struct cmd_binding *getAllBinds(struct cmd_binding *last) {
    int bind_index;
    if(last) {
        if(last->next)
            return last->next;
        bind_index = get_binds_index(last->cmd[0]) + 1;
        if(bind_index > 26)
            return NULL;
    } else
        bind_index = 0;
    do {
        if(cmd_binds[bind_index])
            return cmd_binds[bind_index];
        bind_index++;
    } while(bind_index < 27);
    return NULL;
}

void unregister_module_commands(int module_id) {
    struct cmd_function *cmdfunct, *nextfunct, *prevfunct = NULL;
    int i;
    for(cmdfunct = cmd_functions; cmdfunct; cmdfunct = nextfunct) {
        nextfunct = cmdfunct->next;
        if(cmdfunct->module_id == module_id) {
            for(i = 0; i < 27; i++) {
                struct cmd_binding *cbind, *next, *last = NULL;
                for(cbind = cmd_binds[i]; cbind; cbind = next) {
                    next = cbind->next;
                    if(cmdfunct == cbind->func) {
                        if(last)
                            last->next = cbind->next;
                        else
                            cmd_binds[i] = cbind->next;
                        free(cbind->cmd);
                        if(cbind->paramcount) {
                            int j;
                            for(j = 0; j < cbind->paramcount; j++)
                                free(cbind->parameters[j]);
                        }
                        if(cbind->channel_access)
                            free(cbind->channel_access);
                        free(cbind);
                    } else
                        last = cbind;
                }
            }
            if(prevfunct) 
                prevfunct->next = nextfunct;
            else
                cmd_functions = nextfunct;
            free(cmdfunct->name);
            free(cmdfunct);
        } else
            prevfunct = cmdfunct;
    }
    struct trigger_callback *cb, *prevcb = NULL, *nextcb;
    for(cb = trigger_callbacks; cb; cb = nextcb) {
        nextcb = cb->next;
        if(cb->module_id == module_id) {
            if(prevcb)
                prevcb->next = nextcb;
            else
                trigger_callbacks = nextcb;
            free(cb);
        } else
            prevcb = cb;
    }
}
